--SOUNDBANK SYSTEM V0.3
--CREATED BY DUMMIESMAN
--LATEST REVISION DEC 31 2014

local M = {}
local loadedSounds = {}
local windSound = nil
local cameraNode = 0


--Debug stuff
local wheelSounds = true
local windSounds = true
local windSoundScale = 1
local wheelSoundScale = 1
local wheelRollSoundScale = 1

--The sound profiles. Warning : LOONG CODE :)
--Type enumeration
--Type 0 : Uses engine 
--Type 1 : Activated on key hold
--Type 2 : Wind sound
local RANGEHORN = {
	Type	=	1,
	SFXProfile = "RANGEHORN",
	Pitch	=	1,
	Volume	=	1,
	KeyName	=	"H",
	Node 	=	6,
}

local generic_wind = {
	Type = 2,
	SFXProfile = "WindTestSound",
	VolumeFactor = 0.001,
	PitchFactor = 60,
	MaxVolume = 1,
}

local mclaren_idle = {
	Type	=	0,
	SFXProfile	=	"mclaren_idle",
	MinVolume	=	0.719000,
	MaxVolume	=	0.940000,
	FadeInStartRPM	=	1.000000,
	FadeInEndRPM	=	800.000000,
	FadeOutStartRPM	=	2500.000000,
	FadeOutEndRPM	=	9500.000000,
	MinPitch	=	1.000000,
	MaxPitch	=	1.250000,
	PitchShiftStartRPM	=	1.000000,
	PitchShiftEndRPM	=	4000.000000,
}
local mclaren_low = {
	Type	=	0,
	SFXProfile	=	"mclaren_low_on",
	SFXProfileOff = "mclaren_low_off",
	MinVolume	=	0.679000,
	MaxVolume	=	0.950600,
	FadeInStartRPM	=	500.000000,
	FadeInEndRPM	=	2500.000000,
	FadeOutStartRPM	=	4500.000000,
	FadeOutEndRPM	=	11000.000000,
	MinPitch	=	0.850000,
	MaxPitch	=	2.500000,
	PitchShiftStartRPM	=	500.000000,
	PitchShiftEndRPM	=	13000.000000,
}
local mclaren_mid = {
	Type	=	0,
	SFXProfile	=	"mclaren_mid_on",
	SFXProfileOff = "mclaren_mid_off",
	MinVolume	=	0.630500,
	MaxVolume	=	0.940900,
	FadeInStartRPM	=	1500.000000,
	FadeInEndRPM	=	4500.000000,
	FadeOutStartRPM	=	9000.000000,
	FadeOutEndRPM	=	13500.000000,
	MinPitch	=	0.850000,
	MaxPitch	=	2.500000,
	PitchShiftStartRPM	=	3000.000000,
	PitchShiftEndRPM	=	20000.000000,
}
local mclaren_high = {
	Type	=	0,
	SFXProfile	=	"mclaren_high_on",
	SFXProfileOff = "mclaren_high_off",
	MinVolume	=	0.620800,
	MaxVolume	=	0.940900,
	FadeInStartRPM	=	4500.000000,
	FadeInEndRPM	=	9000.000000,
	FadeOutStartRPM	=	12000.000000,
	FadeOutEndRPM	=	15000.000000,
	MinPitch	=	0.850000,
	MaxPitch	=	3.000000,
	PitchShiftStartRPM	=	4500.000000,
	PitchShiftEndRPM	=	22000.000000,
}




--our "bank" of sounds
local soundBank = {mclaren_idle,mclaren_low,mclaren_mid,mclaren_high,generic_wind,RANGEHORN}


 local function clamp01(value)
    if value < 0 then
		return 0
     end 
     if value > 1 then
		return 1
     end
     return value
 end

local function sblerp(from,to,t)
     return from + (to - from) * clamp01(t)
end


local function nop()
end

local wheelsSounds = nil

local function playWheelSound(wheelID, soundName, vol, pitch)
    ws = wheelsSounds[wheelID][soundName]
    if vol < 0.1 then vol = 0 end
    if vol == 0 then 
        pitch = 0
    else
        if math.abs(vol - ws.lastVol) / vol < 0.05 and math.abs(pitch - ws.lastPitch) < 0.05 then return end
    end
    if vol == ws.lastvol then return end
    ws.lastVol = vol
    ws.lastPitch = pitch
    ws.obj:setVolume(vol)
    ws.obj:setPitch(pitch)
end


local throttleThreshold = 0.5

local function cap(val,maxVal)
	if val > maxVal then return maxVal end
	return val
end

local function UpdateEngineVolumes(curThrottle,volume,entry)
	if entry["SFXProfileOff"] == nil then
		--We don't have an off sound
		entry["clip"]:setVolume(volume)
	else
		--On/off blending values
		local onVolumeFactor = cap(curThrottle * 2,1)
		local offVolumeFactor = sblerp(1,0,cap(curThrottle ,1))
		entry["clip"]:setVolume(volume * onVolumeFactor)
		entry["clipOff"]:setVolume(volume * offVolumeFactor)
	end
end

local function ProcessBankEntry(entry)

	if entry["Type"] == 0 then
		local curRPM = drivetrain.rpm;
		local curThrottle = electrics.values['throttle']
		
		

		
		--Engine type
		if entry["clip"] ~= nil and v.data.engine ~= nil then
				--do the pitch stuff
				
				if curRPM > entry["PitchShiftStartRPM"] and curRPM < entry["PitchShiftEndRPM"] then
					local mrpm = entry["PitchShiftEndRPM"] -  entry["PitchShiftStartRPM"]
					local crpm = curRPM - entry["PitchShiftStartRPM"]
					local pct = crpm/mrpm
					local pitch = sblerp (entry["MinPitch"],entry["MaxPitch"],pct)
					entry["clip"]:setPitch(pitch)
					entry["clipOff"]:setPitch(pitch)
					
				elseif curRPM < entry["PitchShiftStartRPM"] then
					entry["clip"]:setPitch(entry["MinPitch"])
					entry["clipOff"]:setPitch(entry["MinPitch"])
				elseif curRPM > entry["PitchShiftEndRPM"] then
					entry["clip"]:setPitch(entry["MaxPitch"])
					entry["clipOff"]:setPitch(entry["MaxPitch"])
				end
				-- do the volume stuff
				if curRPM > entry["FadeInStartRPM"] and curRPM < entry["FadeInEndRPM"]  then
					--We need to fade in
					local fiSRPM = entry["FadeInStartRPM"]
					local percentage = (curRPM - fiSRPM) / (entry["FadeInEndRPM"] - fiSRPM)
					local vol = sblerp (0,entry["MinVolume"],percentage)
					UpdateEngineVolumes(curThrottle,vol,entry)
				elseif curRPM > entry["FadeInEndRPM"] and curRPM < entry["FadeOutStartRPM"] then
					--set volume based on rpm
					local start =    entry["FadeOutStartRPM"] - entry["FadeInEndRPM"]
					local crpm = curRPM - entry["FadeInEndRPM"]
					local vol = sblerp (entry["MinVolume"],entry["MaxVolume"],crpm/start)
					UpdateEngineVolumes(curThrottle,vol,entry)
				elseif curRPM >  entry["FadeOutStartRPM"]  and curRPM < entry["FadeOutEndRPM"]  then
					--We need to fade out
					local mrpm = entry["FadeOutEndRPM"]  -  entry["FadeOutStartRPM"] 
					local crpm = curRPM - entry["FadeOutStartRPM"] 
					local vol = sblerp (entry["MaxVolume"],0,crpm/mrpm)
					UpdateEngineVolumes(curThrottle,vol,entry)
				else
					entry["clip"]:setVolume(0)
					entry["clipOff"]:setVolume(0)
				end
			--p[clip] = nil if statement
		end
	elseif entry["Type"] == 1 then
		--KeyPress type
		if input.keys[entry["KeyName"]] then
			entry["clip"]:setPitch(entry["Pitch"])
			entry["clip"]:setVolume(entry["Volume"])
		else
			entry["clip"]:setPitch(0)
			entry["clip"]:setVolume(0)
		end
	elseif entry["Type"] == 2 then
		--Wind
		if windSounds == true then
			local speed = obj:getVelocity():length() -- speed
			local vol = (speed * speed *  entry["VolumeFactor"])
			local pitch = speed / entry["PitchFactor"]
			if vol > entry["MaxVolume"] then vol = entry["MaxVolume"]  end
			windSound:setVolume(vol * windSoundScale)
			windSound:setPitch(pitch)
		end
	else
		print("Unknown bank entry type")
	end

end

local function update(dt)
     --banks
	 
		
   for k, p in pairs(soundBank) do 
	 ProcessBankEntry(p)
   end 
    


	
	
    -- wheels
	
	if wheelSounds == true then
    for wi,wd in pairs(drivetrain.wheelInfo) do
        local w = wd.obj
        local slip = wd.lastSlip
        local slipEnergy = wd.slipEnergy

        local skidVol = 0
        local skidPitch = 0
        local rollVol = 0
        local rollPitch = 0

        if wd.contactMaterialID1 == 10 and wd.contactMaterialID2 == 4 and wd.contactDepth == 0 then
            if slipEnergy > 6 then
                skidVol = math.min(slipEnergy * 0.005 + 0.3, 1)
                skidPitch = math.max(slip - 3, 0) * 0.05 + 0.8
            else
                local wheelSpeed = math.abs(w.angularVelocity * wd.radius)
                rollVol = math.min(math.sqrt(wheelSpeed * 0.018), 1)
                rollPitch = wheelSpeed * 0.125
            end
        end
        
        playWheelSound(wi, "SkidTestSound", skidVol * wheelSoundScale, skidPitch)
        playWheelSound(wi, "RollingTestSound", rollVol * wheelRollSoundScale, rollPitch)
    end
	
	end
end

local function addWheelSound(wheelID, wd, soundName)
    if wheelsSounds == nil then wheelsSounds = {} end
    if wheelsSounds[wheelID] == nil then wheelsSounds[wheelID] = {} end
    
    local s = gameEngine:createSFXSource(soundName)
    if s == nil then 
        return nil 
    end
    s:setVolume(0)
    s:setPitch(1)
    s:play(-1)
    obj:attachSFXNode(s, wd.node1)
    wheelsSounds[wheelID][soundName] = {obj = s, lastVol = 0, lastPitch = 0}
    table.insert(loadedSounds, s)
end



local function createSFXSource(SFXname,nodeNumber)
    local var = gameEngine:createSFXSource(SFXname,nodeNumber)
    if var == nil then 
        return nil 
    end
    var:setVolume(0)
    var:play(-1)
    obj:attachSFXNode(var,nodeNumber or cameraNode)
    table.insert(loadedSounds, var)
    return var
end

local function init_replaced()
		if v.data.camerasInternal ~= nil then
			local k, v = next(v.data.camerasInternal)
			cameraNode = v.camNodeID
		end

    
		for k, p in pairs(soundBank) do
			local newSound = createSFXSource(p["SFXProfile"],p["Node"] or nil)
			p["clip"] = newSound
			if p["Type"] == 0 then
				--Initialize our clipOff variable. Specific to type 0
				p["clipOff"] = createSFXSource(p["SFXProfileOff"] or p["SFXProfile"],p["Node"] or nil)
				if p["clipOff"] ~= nil then
					p["clipOff"]:setVolume(0)
					p["clipOff"]:setPitch(0)
				end
			end
			if p["clip"] ~= nil then
				p["clip"]:setVolume(0)
				p["clip"]:setPitch(0)
			end
		end
    

    if windSound == nil then
        windSound = createSFXSource("WindTestSound",nil)
    end    

    if wheelsSounds == nil then
        for wi,wd in pairs(drivetrain.wheels) do
            addWheelSound(wi, wd, "RollingTestSound")
            addWheelSound(wi, wd, "SkidTestSound")
        end
    end
end

local function destroy()
    for k,v in pairs(loadedSounds) do
        if v and v:isValid() then
            gameEngine:deleteSFXSource(v, true)
            v:destroy()
        end
    end
    loadedSounds = {}
    wheelsSounds = nil
    engineSound = nil
    windSound = nil
end

local function init()
	print("Initializing SoundBank System v0.3")
	
    init_replaced()
    --sounds.destroy()
    sounds.update = update
    sounds.destroy = destroy
	
	print("SoundBank System v0.3 Initialized")
end

M.destroy = destroy
M.init = init
return M